import json
import logging
from typing import List, Callable, Tuple

from bs4 import BeautifulSoup
from dateutil import parser as dateutil_parser

from nicoclient.html_page.html_page import HtmlPage
from nicoclient.model.playlist import Playlist
from nicoclient.model.video import Video

logger = logging.getLogger("nicoclient")


def preload(html_string: str) -> Tuple[str, str, List[Video]]:
    key_map = {
        'video_id': 'id',
        'view_counter': 'views',
        'mylist_counter': 'likes',
        'first_retrieve': 'timestamp',
        'last_res_body': 'description',
        'length_seconds': 'duration'
    }

    def get_videos():
        for line in [line.strip() for line in html_string.split('\n')]:
            if line.startswith('Mylist.preload'):
                idx_start = line.find('[')
                line = line[idx_start:-2]
                for item in json.loads(line):
                    item_data = item['item_data']
                    item_data["is_accessible"] = True
                    item_data['view_counter'] = int(item_data['view_counter'])
                    item_data['mylist_counter'] = int(item_data['mylist_counter'])
                    for key_old, key_new in key_map.items():
                        item_data[key_new] = item_data[key_old]
                    yield Video(**item_data)

    def get_owner():
        for line in html_string.split('\n'):
            if line.strip().startswith('mylist_owner: { user_id:'):
                return line.split(',')[0].split(':')[-1].strip()

    def get_name() -> str:
        return f"Playlist"  # TODO: use regex

    return get_owner(), get_name(), list(get_videos())


def initial_userpage(html_string: str) -> Tuple[str, str, List[Video]]:
    soup = BeautifulSoup(html_string, features="html.parser")
    element = soup.find(id="js-initial-userpage-data")
    json_string = element.get("data-initial-data")
    json_object = json.loads(json_string)
    nvapis = json_object['nvapi']

    def get_videos():
        for nvapi in nvapis:
            for item in nvapi['body']['data']['mylist']['items']:
                video = item['video']
                yield Video(
                    id=video["id"],
                    uploader_id=video["owner"]["id"],
                    title=video["title"],
                    views=video["count"]["view"],
                    likes=video["count"]["mylist"],
                    duration=video["duration"],
                    timestamp=dateutil_parser.parse(video["registeredAt"]).timestamp(),
                    thumbnail_url=video["thumbnail"]["url"],
                    is_accessible=True,  # assume accessible by default
                    description=video["shortDescription"]
                )

    def get_owner() -> str:
        for nvapi in nvapis:
            return nvapi["body"]["data"]["mylist"]["owner"]["id"]

    def get_name() -> str:
        for nvapi in nvapis:
            return nvapi["body"]["data"]["mylist"]["name"]

    return get_owner(), get_name(), list(get_videos())


class Parser:
    def __init__(self, execute):
        self.execute: Callable[[str], Tuple[str, str, List[Video]]] = execute


_PARSERS = [
    (lambda html_string: "Mylist.preload" in html_string, Parser(preload)),
    (lambda html_string: "js-initial-userpage-data" in html_string, Parser(initial_userpage)),
    # (lambda html_string: '<script type="application/ld+json">' in html_string, Parser(application_ld_json))
]


def get_html_parser(html_string: str) -> Parser:
    for predicate, parser in _PARSERS:
        if predicate(html_string):
            logger.info(f"parser found; parser_name={parser.execute.__name__}")
            return parser

    raise RuntimeError("Cannot find any parsers for the given html string")


class PlaylistPage(HtmlPage):
    playlist: Playlist

    def __init__(self, id: str):
        HtmlPage.__init__(self, url=f"https://www.nicovideo.jp/mylist/{id}")
        self.id = id
        self._parse()

    def _parse(self):
        parser = get_html_parser(self.html_string)
        owner_id, name, videos = parser.execute(self.html_string)
        self.playlist = Playlist(
            id=self.id,
            name=name,
            owner_id=owner_id,
            videos=videos,
            is_monitored=False
        )


def get_playlist(playlist_id: str) -> Playlist:
    return PlaylistPage(id=playlist_id).playlist



from typing import Optional

from pydantic import BaseModel


class VideoThumbnail(BaseModel):
    id: str
    title: str
    views: int
    likes: int
    duration: int  # seconds
    is_accessible: bool
    timestamp: int  # epoch format
    thumbnail_url: str = None


class Video(VideoThumbnail):
    parent_video_id: Optional[str] = None
    description: str



import json
from abc import ABC
from html.parser import HTMLParser

import requests


class HtmlPage(ABC):
    def __init__(self, html_string=None, url=None, headers=None, expected_codes=None):
        assert html_string or url
        self.html_string = html_string
        self.status_code = None

        if html_string is None:
            expected_codes = expected_codes or [200]
            response = requests.get(url=url, headers=headers or dict())
            self.status_code = response.status_code
            if response.status_code in expected_codes:
                self.html_string = str(response.text)
            elif response.status_code == 403:
                raise PageAccessDeniedError(url)
            elif response.status_code == 404:
                raise PageNotFoundError(url)
            else:
                a_status = response.status_code
                e_status = json.dumps(expected_codes)
                text = str(response.text)
                raise RuntimeError(f"status_code={a_status} expected_codes={e_status} text='{text}'")


class PageError(Exception):
    pass


class PageAccessDeniedError(PageError):
    pass


class PageNotFoundError(PageError):
    pass


def to_json(content, raise_exception=False):
    def recursive_clean(node):
        if isinstance(node, list):
            for child in node:
                recursive_clean(child)
        elif isinstance(node, dict):
            if '__parent__' in node:
                del node['__parent__']
            for child in node.values():
                recursive_clean(child)

    class HTMLtoJSONParser(HTMLParser):
        def __init__(self, raise_exception=True):
            HTMLParser.__init__(self)
            self.doc = {}
            self.path = []
            self.cur = self.doc
            self.line = 0
            self.raise_exception = raise_exception

        @property
        def json(self):
            return self.doc

        def handle_starttag(self, tag, attrs):
            self.path.append(tag)
            attrs = {k: v for k, v in attrs}
            if tag in self.cur:
                if isinstance(self.cur[tag], list):
                    self.cur[tag].append({"__parent__": self.cur})
                    self.cur = self.cur[tag][-1]
                else:
                    self.cur[tag] = [self.cur[tag]]
                    self.cur[tag].append({"__parent__": self.cur})
                    self.cur = self.cur[tag][-1]
            else:
                self.cur[tag] = {"__parent__": self.cur}
                self.cur = self.cur[tag]

            for a, v in attrs.items():
                self.cur["#" + a] = v
            self.cur[""] = ""

        def handle_endtag(self, tag):
            if tag != self.path[-1] and self.raise_exception:
                raise Exception(
                    "html is malformed around line: {0} (it might be because of a tag <br>, <hr>, <img .. > not closed)".format(
                        self.line))
            del self.path[-1]
            memo = self.cur
            self.cur = self.cur["__parent__"]
            self.clean(memo)

        def handle_data(self, data):
            self.line += data.count("\n")
            if "" in self.cur:
                self.cur[""] += data

        def clean(self, values):
            keys = list(values.keys())
            for k in keys:
                v = values[k]
                if isinstance(v, str):
                    # print ("clean", k,[v])
                    c = v.strip(" \n\r\t")
                    if c != v:
                        if len(c) > 0:
                            values[k] = c
                        else:
                            del values[k]
            del values["__parent__"]

    parser = HTMLtoJSONParser(raise_exception=raise_exception)
    parser.feed(content)
    recursive_clean(parser.json)
    return parser.json



import datetime
import logging
from typing import List, Iterator

import dateutil.parser
import requests

from nicoclient.model.video import Video

_SEARCH_LIMIT = 25

logger = logging.getLogger("nicoclient")
from datetime import datetime, timedelta


class NicoSearchApiV2Client:
    """
    https://site.nicovideo.jp/search-api-docs/snapshot
    """

    @staticmethod
    def search(query: str, param_override: dict = None) -> Iterator[Video]:
        url = "https://api.search.nicovideo.jp/api/v2/snapshot/video/contents/search"
        headers = {"User-Agent": "nicoclient"}
        params = dict(
            q=query,
            targets="title,description,tags",
            fields="contentId,title,description,userId,mylistCounter,viewCounter,lengthSeconds,thumbnailUrl,startTime",
            _sort="-viewCounter",
            _limit=_SEARCH_LIMIT,
            _context="nicoclient"
        )
        if param_override:
            params.update(param_override)

        response = requests.request("GET", url, headers=headers, params=params)
        response.raise_for_status()

        for item in response.json()["data"]:
            yield Video(
                id=item["contentId"],
                likes=item["mylistCounter"],
                duration=item["lengthSeconds"],
                views=item["viewCounter"],
                thumbnail_url=item["thumbnailUrl"],
                description=item["description"],
                timestamp=dateutil.parser.parse(item["startTime"]).timestamp(),
                uploader_id=item["userId"],
                title=item["title"],
                is_accessible=True
            )

    @staticmethod
    def search_daily_trending(min_likes=0):
        yesterday = datetime.now() - timedelta(1)
        return list(NicoSearchApiV2Client.search(
            query="歌ってみた",
            param_override={
                "targets": "tagsExact",
                "filters[startTime][gte]": yesterday.strftime("%Y-%m-%dT%H:%M:%S+00:00"),
                "filters[mylistCounter][gte]": min_likes
            }
        ))


def search_utattemita_videos(video_id: str) -> List[Video]:
    return list(NicoSearchApiV2Client.search(f"{video_id} 歌ってみた"))


search_daily_trending_utatttemita_videos = NicoSearchApiV2Client.search_daily_trending



import dateutil.parser
import requests
from bs4 import BeautifulSoup

from nicoclient.model.video import Video


class Accessor(BeautifulSoup):
    def __init__(self, content):
        super().__init__(content, "html.parser")

    def __getitem__(self, key):
        return self.find(key).text


def to_duration(length: str) -> int:
    split = length.split(":")
    assert len(split) == 2
    min, sec = split
    return int(min) * 60 + int(sec)


def get_video_by_id(video_id) -> Video:
    url = f"https://ext.nicovideo.jp/api/getthumbinfo/{video_id}"
    r = requests.get(url)
    r.raise_for_status()

    accessor = Accessor(r.text)

    return Video(
        id=accessor["video_id"],
        title=accessor["title"],
        description=accessor["description"],
        views=int(accessor["view_counter"]),
        likes=int(accessor["mylist_counter"]),
        thumbnail_url=accessor["thumbnail_url"],
        duration=to_duration(accessor["length"]),
        timestamp=dateutil.parser.parse(accessor["first_retrieve"]).timestamp(),
        is_accessible=True,
    )



from nicoclient.core.lookup import get_video_by_id
from nicoclient.core.search import search_utattemita_videos, search_daily_trending_utatttemita_videos
from nicoclient.html_page.playlist import get_playlist

name = "nicoclient"



from typing import List, Optional

from pydantic import BaseModel

from nicoclient.model.video import VideoThumbnail


class Playlist(BaseModel):
    id: str
    name: str
    owner_id: Optional[str]
    videos: List[VideoThumbnail]
    is_monitored: bool



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nicoclient",
    version="v1.1.7",
    author="Jaeseo Park",
    author_email="jaeseopark@icloud.com",
    description="A python client to interact with nicovideo.jp",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jaeseopark/nicoclient",
    packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
    install_requires=['requests', 'beautifulsoup4', 'python-dateutil', 'pydantic'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Natural Language :: English"
    ]
)



import time
from datetime import datetime


def get_posix_now():
    return get_posix(datetime.now())


def get_posix(datime_obj):
    utc = datime_obj.utctimetuple()
    t = time.mktime(utc)
    return int(t)


def str_to_posix(date_str):
    """
    Converts datetime_str to datetime_int

    :param date_str: A string representation of datetime. The format is "YYYY/mm/DD HH:MM" (16 characters).
                     If the string is bigger than 16 characters long, then it gets truncated.
    :return: A posix/integer representation of datetime
    """
    dt = datetime.strptime(date_str[:16], "%Y/%m/%d %H:%M")
    return get_posix(dt)
